@@ -22,22 +22,37 @@ typedef struct {
     char *val;
 } zend_string;
 
+#define REDIS_MAKE_STD_ZVAL(zv) MAKE_STD_ZVAL(zv)
+#define REDIS_FREE_ZVAL(zv) (efree(zv))
+
 #define ZSTR_VAL(s) (s)->val
 #define ZSTR_LEN(s) (s)->len
 
 static zend_always_inline zend_string *
-zend_string_init(const char *str, size_t len, int persistent)
+zend_string_alloc(size_t len, int persistent)
 {
-    zend_string *zstr = emalloc(sizeof(zend_string) + len + 1);
+    zend_string *zstr = emalloc(sizeof(*zstr) + len + 1);
 
-    ZSTR_VAL(zstr) = (char *)zstr + sizeof(zend_string);
-    memcpy(ZSTR_VAL(zstr), str, len);
-    ZSTR_VAL(zstr)[len] = '\0';
+    ZSTR_VAL(zstr) = (char *)zstr + sizeof(*zstr);
     zstr->len = len;
     zstr->gc = 0x01;
     return zstr;
 }
 
+static zend_always_inline zend_string *
+zend_string_init(const char *str, size_t len, int persistent)
+{
+    zend_string *zstr = zend_string_alloc(len, persistent);
+
+    memcpy(ZSTR_VAL(zstr), str, len);
+    ZSTR_VAL(zstr)[len] = '\0';
+    return zstr;
+}
+
+#define zend_string_equal_val(s1, s2) !memcmp(ZSTR_VAL(s1), ZSTR_VAL(s2), ZSTR_LEN(s1))
+#define zend_string_equal_content(s1, s2) (ZSTR_LEN(s1) == ZSTR_LEN(s2) && zend_string_equal_val(s1, s2))
+#define zend_string_equals(s1, s2) (s1 == s2 || zend_string_equal_content(s1, s2))
+
 #define zend_string_release(s) do { \
     if ((s) && (s)->gc) { \
         if ((s)->gc & 0x10 && ZSTR_VAL(s)) efree(ZSTR_VAL(s)); \
@@ -412,11 +427,17 @@ typedef int strlen_t;
 /* References don't need any actions */
 #define ZVAL_DEREF(v) PHPREDIS_NOTUSED(v)
 
+#define PHPREDIS_GET_OBJECT(class_entry, z) (class_entry *)zend_objects_get_address(z TSRMLS_CC)
+
 #else
 #include <zend_smart_str.h>
 #include <ext/standard/php_smart_string.h>
 typedef size_t strlen_t;
 #define PHPREDIS_ZVAL_IS_STRICT_FALSE(z) (Z_TYPE_P(z) == IS_FALSE)
+#define PHPREDIS_GET_OBJECT(class_entry, z) (class_entry *)((char *)Z_OBJ_P(z) - XtOffsetOf(class_entry, std))
+
+#define REDIS_MAKE_STD_ZVAL(zv) do {} while(0)
+#define REDIS_FREE_ZVAL(zv) do {} while(0)
 #endif
 
 /* NULL check so Eclipse doesn't go crazy */
@@ -440,8 +461,10 @@ typedef size_t strlen_t;
 
 #ifdef PHP_WIN32
 #define PHP_REDIS_API __declspec(dllexport)
+#define phpredis_atoi64(p) _atoi64((p))
 #else
 #define PHP_REDIS_API
+#define phpredis_atoi64(p) atoll((p))
 #endif
 
 /* reply types */
@@ -475,6 +498,8 @@ typedef enum _PUBSUB_TYPE {
 #define REDIS_OPT_READ_TIMEOUT       3
 #define REDIS_OPT_SCAN               4
 #define REDIS_OPT_FAILOVER           5
+#define REDIS_OPT_TCP_KEEPALIVE      6
+#define REDIS_OPT_COMPRESSION        7
 
 /* cluster options */
 #define REDIS_FAILOVER_NONE              0
@@ -485,6 +510,9 @@ typedef enum _PUBSUB_TYPE {
 #define REDIS_SERIALIZER_NONE        0
 #define REDIS_SERIALIZER_PHP         1
 #define REDIS_SERIALIZER_IGBINARY    2
+/* compression */
+#define REDIS_COMPRESSION_NONE 0
+#define REDIS_COMPRESSION_LZF  1
 
 /* SCAN options */
 #define REDIS_SCAN_NORETRY 0
@@ -499,12 +527,9 @@ typedef enum _PUBSUB_TYPE {
 #define MULTI    1
 #define PIPELINE 2
 
-#define IF_ATOMIC() if (redis_sock->mode == ATOMIC)
-#define IF_NOT_ATOMIC() if (redis_sock->mode != ATOMIC)
-#define IF_MULTI() if (redis_sock->mode & MULTI)
-#define IF_NOT_MULTI() if (!(redis_sock->mode & MULTI))
-#define IF_PIPELINE() if (redis_sock->mode & PIPELINE)
-#define IF_NOT_PIPELINE() if (!(redis_sock->mode & PIPELINE))
+#define IS_ATOMIC(redis_sock) (redis_sock->mode == ATOMIC)
+#define IS_MULTI(redis_sock) (redis_sock->mode & MULTI)
+#define IS_PIPELINE(redis_sock) (redis_sock->mode & PIPELINE)
 
 #define PIPELINE_ENQUEUE_COMMAND(cmd, cmd_len) do { \
     if (redis_sock->pipeline_cmd == NULL) { \
@@ -539,7 +564,7 @@ typedef enum _PUBSUB_TYPE {
 } while (0)
 
 #define REDIS_PROCESS_REQUEST(redis_sock, cmd, cmd_len) \
-    IF_PIPELINE() { \
+    if (IS_PIPELINE(redis_sock)) { \
         PIPELINE_ENQUEUE_COMMAND(cmd, cmd_len); \
     } else { \
         SOCKET_WRITE_COMMAND(redis_sock, cmd, cmd_len); \
@@ -547,7 +572,7 @@ typedef enum _PUBSUB_TYPE {
     efree(cmd);
 
 #define REDIS_PROCESS_RESPONSE_CLOSURE(function, closure_context) \
-    IF_NOT_PIPELINE() { \
+    if (!IS_PIPELINE(redis_sock)) { \
         if (redis_response_enqueued(redis_sock TSRMLS_CC) != SUCCESS) { \
             RETURN_FALSE; \
         } \
@@ -575,7 +600,7 @@ typedef enum _PUBSUB_TYPE {
             RETURN_FALSE; \
     } \
     REDIS_PROCESS_REQUEST(redis_sock, cmd, cmd_len); \
-    IF_ATOMIC() { \
+    if (IS_ATOMIC(redis_sock)) { \
         resp_func(INTERNAL_FUNCTION_PARAM_PASSTHRU, redis_sock, NULL, ctx); \
     } else { \
         REDIS_PROCESS_RESPONSE_CLOSURE(resp_func, ctx) \
@@ -591,19 +616,12 @@ typedef enum _PUBSUB_TYPE {
             RETURN_FALSE; \
     } \
     REDIS_PROCESS_REQUEST(redis_sock, cmd, cmd_len); \
-    IF_ATOMIC() { \
+    if (IS_ATOMIC(redis_sock)) { \
         resp_func(INTERNAL_FUNCTION_PARAM_PASSTHRU, redis_sock, NULL, ctx); \
     } else { \
         REDIS_PROCESS_RESPONSE_CLOSURE(resp_func, ctx) \
     }
 
-#define REDIS_STREAM_CLOSE_MARK_FAILED(redis_sock) \
-    redis_stream_close(redis_sock TSRMLS_CC); \
-    redis_sock->stream = NULL; \
-    redis_sock->mode = ATOMIC; \
-    redis_sock->status = REDIS_SOCK_STATUS_FAILED; \
-    redis_sock->watching = 0
-
 /* Extended SET argument detection */
 #define IS_EX_ARG(a) \
     ((a[0]=='e' || a[0]=='E') && (a[1]=='x' || a[1]=='X') && a[2]=='\0')
@@ -648,6 +666,7 @@ typedef struct {
     zend_string    *persistent_id;
 
     int            serializer;
+    int            compression;
     long           dbNumber;
 
     zend_string    *prefix;
@@ -661,11 +680,10 @@ typedef struct {
 
     zend_string    *err;
 
-    zend_bool      lazy_connect;
-
     int            scan;
 
     int            readonly;
+    int            tcp_keepalive;
 } RedisSock;
 /* }}} */
 
@@ -775,6 +793,11 @@ ZEND_BEGIN_ARG_INFO_EX(arginfo_key_offset_value, 0, 0, 3)
     ZEND_ARG_INFO(0, value)
 ZEND_END_ARG_INFO()
 
+ZEND_BEGIN_ARG_INFO_EX(arginfo_swapdb, 0, 0, 2)
+    ZEND_ARG_INFO(0, srcdb)
+    ZEND_ARG_INFO(0, dstdb)
+ZEND_END_ARG_INFO()
+
 ZEND_BEGIN_ARG_INFO_EX(arginfo_key_start_end, 0, 0, 3)
     ZEND_ARG_INFO(0, key)
     ZEND_ARG_INFO(0, start)
@@ -793,8 +816,7 @@ ZEND_END_ARG_INFO()
 ZEND_BEGIN_ARG_INFO_EX(arginfo_set, 0, 0, 2)
     ZEND_ARG_INFO(0, key)
     ZEND_ARG_INFO(0, value)
-    ZEND_ARG_INFO(0, timeout)
-    ZEND_ARG_INFO(0, opt)
+    ZEND_ARG_INFO(0, opts)
 ZEND_END_ARG_INFO()
 
 ZEND_BEGIN_ARG_INFO_EX(arginfo_lset, 0, 0, 3)
@@ -1015,10 +1037,6 @@ ZEND_BEGIN_ARG_INFO_EX(arginfo_watch, 0, 0, 1)
 #endif
 ZEND_END_ARG_INFO()
 
-ZEND_BEGIN_ARG_INFO_EX(arginfo_info, 0, 0, 0)
-    ZEND_ARG_INFO(0, option)
-ZEND_END_ARG_INFO()
-
 ZEND_BEGIN_ARG_INFO_EX(arginfo_command, 0, 0, 0)
 #if PHP_VERSION_ID >= 50600
     ZEND_ARG_VARIADIC_INFO(0, args)
@@ -1036,44 +1054,6 @@ ZEND_BEGIN_ARG_INFO_EX(arginfo_rawcommand, 0, 0, 1)
 #endif
 ZEND_END_ARG_INFO()
 
-ZEND_BEGIN_ARG_INFO_EX(arginfo_client, 0, 0, 1)
-    ZEND_ARG_INFO(0, cmd)
-#if PHP_VERSION_ID >= 50600
-    ZEND_ARG_VARIADIC_INFO(0, args)
-#else
-    ZEND_ARG_INFO(0, ...)
-#endif
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO_EX(arginfo_config, 0, 0, 2)
-    ZEND_ARG_INFO(0, cmd)
-    ZEND_ARG_INFO(0, key)
-    ZEND_ARG_INFO(0, value)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO_EX(arginfo_pubsub, 0, 0, 1)
-    ZEND_ARG_INFO(0, cmd)
-#if PHP_VERSION_ID >= 50600
-    ZEND_ARG_VARIADIC_INFO(0, args)
-#else
-    ZEND_ARG_INFO(0, ...)
-#endif
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO_EX(arginfo_script, 0, 0, 1)
-    ZEND_ARG_INFO(0, cmd)
-#if PHP_VERSION_ID >= 50600
-    ZEND_ARG_VARIADIC_INFO(0, args)
-#else
-    ZEND_ARG_INFO(0, ...)
-#endif
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO_EX(arginfo_slowlog, 0, 0, 1)
-    ZEND_ARG_INFO(0, arg)
-    ZEND_ARG_INFO(0, option)
-ZEND_END_ARG_INFO()
-
 ZEND_BEGIN_ARG_INFO_EX(arginfo_geoadd, 0, 0, 4)
     ZEND_ARG_INFO(0, key)
     ZEND_ARG_INFO(0, lng)
@@ -1110,4 +1090,82 @@ ZEND_BEGIN_ARG_INFO_EX(arginfo_georadiusbymember, 0, 0, 4)
     ZEND_ARG_ARRAY_INFO(0, opts, 0)
 ZEND_END_ARG_INFO()
 
+ZEND_BEGIN_ARG_INFO_EX(arginfo_xadd, 0, 0, 3)
+    ZEND_ARG_INFO(0, str_key)
+    ZEND_ARG_INFO(0, str_id)
+    ZEND_ARG_ARRAY_INFO(0, arr_fields, 0)
+    ZEND_ARG_INFO(0, i_maxlen)
+    ZEND_ARG_INFO(0, boo_approximate)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(arginfo_xpending, 0, 0, 2)
+    ZEND_ARG_INFO(0, str_key)
+    ZEND_ARG_INFO(0, str_group)
+    ZEND_ARG_INFO(0, str_start)
+    ZEND_ARG_INFO(0, str_end)
+    ZEND_ARG_INFO(0, i_count)
+    ZEND_ARG_INFO(0, str_consumer)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(arginfo_xrange, 0, 0, 3)
+    ZEND_ARG_INFO(0, str_key)
+    ZEND_ARG_INFO(0, str_start)
+    ZEND_ARG_INFO(0, str_end)
+    ZEND_ARG_INFO(0, i_count)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(arginfo_xread, 0, 0, 1)
+    ZEND_ARG_ARRAY_INFO(0, arr_streams, 0)
+    ZEND_ARG_INFO(0, i_count)
+    ZEND_ARG_INFO(0, i_block)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(arginfo_xreadgroup, 0, 0, 3)
+    ZEND_ARG_INFO(0, str_group)
+    ZEND_ARG_INFO(0, str_consumer)
+    ZEND_ARG_ARRAY_INFO(0, arr_streams, 0)
+    ZEND_ARG_INFO(0, i_count)
+    ZEND_ARG_INFO(0, i_block)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(arginfo_xack, 0, 0, 3)
+    ZEND_ARG_INFO(0, str_key)
+    ZEND_ARG_INFO(0, str_group)
+    ZEND_ARG_ARRAY_INFO(0, arr_ids, 0)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(arginfo_xclaim, 0, 0, 5)
+    ZEND_ARG_INFO(0, str_key)
+    ZEND_ARG_INFO(0, str_group)
+    ZEND_ARG_INFO(0, str_consumer)
+    ZEND_ARG_INFO(0, i_min_idle)
+    ZEND_ARG_ARRAY_INFO(0, arr_ids, 0)
+    ZEND_ARG_ARRAY_INFO(0, arr_opts, 0)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(arginfo_xgroup, 0, 0, 1)
+    ZEND_ARG_INFO(0, str_operation)
+    ZEND_ARG_INFO(0, str_key)
+    ZEND_ARG_INFO(0, str_arg1)
+    ZEND_ARG_INFO(0, str_arg2)
+    ZEND_ARG_INFO(0, str_arg3)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(arginfo_xinfo, 0, 0, 1)
+    ZEND_ARG_INFO(0, str_cmd)
+    ZEND_ARG_INFO(0, str_key)
+    ZEND_ARG_INFO(0, str_group)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(arginfo_xtrim, 0, 0, 2)
+    ZEND_ARG_INFO(0, str_key)
+    ZEND_ARG_INFO(0, i_maxlen)
+    ZEND_ARG_INFO(0, boo_approximate)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(arginfo_xdel, 0, 0, 2)
+    ZEND_ARG_INFO(0, str_key)
+    ZEND_ARG_ARRAY_INFO(0, arr_ids, 0)
+ZEND_END_ARG_INFO()
+
 #endif
